"""
This file is part of the tagup Python module which is released under MIT.
See file LICENSE for full license details.
"""


from pkg_resources import get_distribution, DistributionNotFound

from .language import Renderer


# Set version number.
try:
	__version__ = get_distribution(__name__).version
except DistributionNotFound:
	__version__ = ""



"""
This file is part of the tagup Python module which is released under MIT.
See file LICENSE for full license details.
"""


from copy import deepcopy

from lark import Token, Tree

from .traversal import (
	DiscardNode,
	PostOrderTraverser,
	PreOrderTraverser
)


class ContextMixin:
	def __init__(self, named_args, pos_args, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self.named_args = named_args
		self.pos_args = pos_args


class CommonEvaluator(ContextMixin, PostOrderTraverser):
	escape_sequences = {
		'o': '[',
		'c': ']',
		's': '\\',
	}

	def __init__(self, renderer, trim_args, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self.renderer = renderer
		self.trim_args = trim_args

	def escape_sequence(self, node):
		sequence = node.children[0]

		return self.escape_sequences[sequence]

	def named_substitution(self, node):
		name = node.children[0]

		return self.named_args[name.trim()]

	def positional_substitution(self, node):
		position = node.children[0]
		arg_num = int(position.trim()) - 1

		return self.pos_args[arg_num]

	def block(self, node):
		children = node.children

		return ''.join(children)

	def named_argument(self, node):
		name = node.children[0]
		value = node.children[1]
		if self.trim_args:
			value = value.trim()

		return (name, value)

	def positional_argument(self, node):
		value = node.children[0]
		if self.trim_args:
			value = value.trim()

		return (value,)

	def tag(self, node):
		children = node.children
		name = children[0]
		named_args = dict()
		pos_args = list()
		for arg in children[1:]:
			if len(arg) == 2:
				named_args[arg[0]] = arg[1]
			else:
				pos_args.append(arg[0])

		return self.renderer._render_tag(
			name=name,
			named_args=named_args,
			pos_args=pos_args
		)


class ControlFlowEvaluator(ContextMixin, PreOrderTraverser):
	def named_test(self, node):
		children = node.children
		name = children[0]
		else_clause = len(children) == 3
		if name in self.named_args:
			result = children[1].children[0]
		elif else_clause:
			result = children[2].children[0]
		else:
			raise DiscardNode()

		return result

	def positional_test(self, node):
		children = node.children
		arg_num = int(children[0]) - 1
		else_clause = len(children) == 3
		if len(self.pos_args) > arg_num:
			result = children[1].children[0]
		elif else_clause:
			result = children[2].children[0]
		else:
			raise DiscardNode()

		return result

	def positional_loop(self, node):
		children = node.children
		else_clause = len(children) == 2
		if len(self.pos_args) > 0:
			statement = children[0].children[0]
			result = Tree(data='block', children=list())
			for arg in self.pos_args:
				new_statement = deepcopy(statement)
				sub_args = new_statement.find_data('loop_item')
				for sub_arg in sub_args:
					sub_arg.data = 'block'
					sub_arg.children = [
						Token('STRING', arg),
					]
				result.children.append(new_statement)
		elif else_clause:
			result = children[1].children[0]
		else:
			raise DiscardNode()

		return result



"""
This file is part of the tagup Python module which is released under MIT.
See file LICENSE for full license details.
"""


from os import path
from lark import Lark, Tree

from .evaluation import CommonEvaluator, ControlFlowEvaluator


class Renderer:
	def __init__(
		self,
		get_tag_callback,
		cache_tag_ast_callback=None,
		trim_args=False
	):
		grammar = self._get_grammar()
		self._parser = Lark(grammar, parser='lalr')
		self._get_tag = get_tag_callback
		self._cache_tag_ast = cache_tag_ast_callback
		self._trim_args = trim_args

	def render_markup(self, markup, named_args=dict(), pos_args=list()):
		ast = self._parse_markup(markup)
		result = self._evaluate_ast(ast, named_args, pos_args)

		return result

	def _render_tag(self, name, named_args, pos_args):
		tag = self._get_tag(name)
		if isinstance(tag, Tree):
			ast = tag
		else:
			ast = self._parse_markup(tag)
			if self._cache_tag_ast is not None:
				self._cache_tag_ast(name, ast)

		return self._evaluate_ast(ast, named_args, pos_args)

	def _parse_markup(self, markup):
		return self._parser.parse(markup)

	def _evaluate_ast(self, ast, named_args, pos_args):
		intermediate = ControlFlowEvaluator(
			named_args=named_args,
			pos_args=pos_args,
		).traverse(ast)
		result = CommonEvaluator(
			named_args=named_args,
			pos_args=pos_args,
			renderer=self,
			trim_args=self._trim_args
		).traverse(intermediate)

		return result

	def _get_grammar(self):
		root_dir = path.dirname(path.abspath(__file__))
		grammar_file = path.join(root_dir, 'grammar.lark')
		with open(grammar_file) as f_in:
			return f_in.read()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tagup']

package_data = \
{'': ['*']}

install_requires = \
['lark-parser>=0.8.5,<0.9.0']

setup_kwargs = {
    'name': 'tagup',
    'version': '0.1.1',
    'description': 'Reference implementation of the Tagup Language',
    'long_description': '# tagup\n\n**tagup** is a Python module which provides a reference implementation of the [Tagup Language](https://fairburn.dev/tagup/).\n\nThis module currently implements [version 1.0.0](https://fairburn.dev/tagup/1.0.0/) of the Tagup Language.\n\n## Changelog\n\n**v0.1.1**\n\n- Added non-standard option to trim whitespace from arguments before tag evaluation.\n- Fixed bug where whitespace was considered when specifying a name/position for argument substitution.\n\n**v0.1.0**\n\n- Initial release.\n',
    'author': 'Garrett Fairburn',
    'author_email': 'garrett@fairburn.dev',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://fairburn.dev/tagup/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3,<4',
}


setup(**setup_kwargs)



"""
This file is part of the tagup Python module which is released under MIT.
See file LICENSE for full license details.
"""


from lark import Tree


class DiscardNode(Exception):
	pass


class PostOrderTraverser:
	def traverse(self, node):
		f = getattr(self, node.data)
		node.children = self._traverse_children(node)
		new_node = f(node)

		return new_node

	def _traverse_children(self, node):
		new_children = []
		for child in node.children:
			if isinstance(child, Tree):
				try:
					new_children.append(self.traverse(child))
				except DiscardNode as e:
					pass
			else:
				new_children.append(child)

		return new_children

	def __getattr__(self, name):
		return self.__default__

	def __default__(self, node):
		return node


class PreOrderTraverser(PostOrderTraverser):
	def traverse(self, node):
		f = getattr(self, node.data)
		new_node = f(node)
		new_node.children = self._traverse_children(new_node)

		return new_node
